"""Type definitions."""

# Future Imports
from __future__ import annotations

# Standard Library Imports
import typing

__all__ = (
    "AnyCallable",
    "AnyCoroutine",
    "AnyException",
)


AnyCallable = typing.Callable[..., typing.Any]
AnyException = typing.Union[Exception, typing.Type[Exception]]
AnyCoroutine = typing.Coroutine[typing.Any, typing.Any, typing.Any]



"""Custom Celery worker pool."""

# Future Imports
from __future__ import annotations

# Standard Library Imports
import asyncio as aio
import inspect
import os
import sys
import threading
import time
from typing import (
    Any,
    Callable,
    Optional,
    Union,
)

# Third-Party Imports
import celery.concurrency.base
import celery.concurrency.solo
import celery.signals
from billiard.einfo import ExceptionInfo
from billiard.exceptions import WorkerLostError
from celery.exceptions import (
    WorkerShutdown,
    WorkerTerminate,
    reraise,
)

# Package-Level Imports
from celery_aio_pool.types import (
    AnyCallable,
    AnyCoroutine,
    AnyException,
)

__all__ = ("AsyncIOPool",)


WorkerPoolInfo = dict[
    str,
    Optional[
        Union[
            int,
            bool,
            tuple[int, ...],
            aio.AbstractEventLoop,
        ]
    ],
]


class AsyncIOPool(celery.concurrency.solo.TaskPool):
    """Custom asyncio Celery worker pool class."""

    loop: aio.AbstractEventLoop
    loop_runner: threading.Thread
    singleton: Optional["AsyncIOPool"] = None

    def __new__(cls, *args: Any, **kwargs: Any) -> "AsyncIOPool":

        # Because the worker pool uses an instance-bound thread
        # to run its asyncio eventloop, it's a good idea to ensure
        # that there can't be more than one instance of the pool
        # created per process

        if not isinstance(cls.singleton, cls):
            # Package-Level Imports
            from celery_aio_pool import patch_celery_tracer

            # We can't assume that a user has elected to enable
            # automatic patching of Celery's default `build_tracer`
            # utility. We need to be *guarantee* that the patch
            # has been applied though, and the patching process
            # itself is idempotent therefore safe to call any
            # number of times.
            assert patch_celery_tracer()

            # Create the requested new worker pool and use it to
            # populate the class's currently "empty" `singleton`
            # attribute
            cls.singleton = super(AsyncIOPool, cls).__new__(cls)

        # Return the class-bound worker pool instance
        return cls.singleton

    def __init__(self, *args: Any, **kwargs: Any) -> None:

        # If there is already a running asyncio event loop
        # in the current thread / process, it's not a great
        # idea to allow the worker pool to create another
        # and the `AsyncIOPool` can't function at all without
        # one, so we'll throw a `SystemError` if one is detected
        # as a way to short-circuit the worker start up process
        try:
            aio.get_running_loop()
            raise SystemError("There is already a running event loop in this thread!")
        except RuntimeError:
            pass

        # Regardless of what the user specifies in the local
        # configuration, `threads` and `forking_enable` should
        # *always* be False when using `AsyncIOPool` as the
        # worker pool class, so we'll set them forcibly here

        kwargs.update(
            threads=False,
            forking_enable=False,
        )

        # Call the default constructor method ...
        celery.concurrency.base.BasePool.__init__(
            self,
            *args,
            **kwargs,
        )

        # ... perform the usual "housekeeping", ...
        self.limit = 1
        celery.signals.worker_process_init.send(sender=None)

        # ... create the pool's asyncio eventloop ...
        self.loop = aio.new_event_loop()

        # ... and let it run in an instance-bound thread.
        self.loop_runner = threading.Thread(
            target=self.loop.run_forever,
            name="celery-worker-async-loop",
            daemon=True,
        )

        self.loop_runner.start()

        # Set the new event loop as the "active" eventloop
        # in current thread / process
        aio.set_event_loop(self.loop)

    def _get_info(self) -> WorkerPoolInfo:
        info = super()._get_info()
        info.update({
            "timeouts": (),
            "max-concurrency": 1,
            "event-loop": str(self.loop),
            "max-tasks-per-child": None,
            "processes": (os.getpid(),),
            "put-guarded-by-semaphore": True,
        })
        return info

    def run(
        self,
        task_function: AnyCallable | AnyCoroutine,
        *args: Any,
        **kwargs: Any,
    ) -> Any:
        """Run the supplied coroutine in the pool's bound loop-runner
        thread."""

        # If the supplied function is actually an async function
        # (i.e. async def some_function() -> Any), call it with
        # the supplied arguments and bind the returned coroutine
        # so we can run it on the worker's thread-bound eventloop
        if inspect.iscoroutinefunction(task_function):
            task_function = task_function(*args, **kwargs)

        # If the supplied function is actually a vanilla Python
        # function (i.e. def any_function() -> Any), use asyncio's
        # `to_thread` utility to wrap it along the supplied arguments
        # and bind the returned coroutine so we can run it on the
        # worker's thread-bound eventloop
        if callable(task_function) and not bool(
            inspect.iscoroutine(task_function) or aio.isfuture(task_function)
        ):
            task_function = aio.to_thread(
                task_function,
                *args,
                **kwargs,
            )

        # If the value of `task_function` isn't actually something
        # that can be awaited (i.e. run on an async eventloop),
        # return it as it's either the actual result of having
        # called and run `task_function` -or- it's something we
        # can't meaningfully do anything with anyway
        if not inspect.isawaitable(task_function):
            return task_function

        # At this point, we're guaranteed to have something
        # that's either an actual coroutine or some other kind
        # of `asyncio.Future` which means we need to throw it
        # onto the worker's thread-bound eventloop to be run
        result: aio.Future = aio.run_coroutine_threadsafe(
            task_function,
            self.loop,
        )

        # Once the our future has been awaited, it will either
        # have raised an exception or returned a result. If it
        # raised an exception, propagate it back to the caller
        if error := result.exception():
            raise error

        # If no exception was raised, pass `.result()` back through
        # `AsyncIOPool.run` so that it can be checked to ensure it's
        # properly handled in case another callable or awaitable was
        # returned
        return self.run(result.result())

    @classmethod
    def run_in_pool(
        cls,
        task_function: AnyCallable | AnyCoroutine,
        *args: Any,
        **kwargs: Any,
    ) -> Any:
        """Run the supplied task in the pool's thread-bound async loop."""
        if not (worker_pool := cls.singleton):
            worker_pool = cls()

        return worker_pool.run(
            task_function,
            *args,
            **kwargs,
        )

    async def shutdown(self) -> None:
        """Shut down the worker pool."""
        if self.loop.is_running():
            self.loop.stop()
            await self.loop.shutdown_asyncgens()

        if not self.loop.is_closed() and callable(
            (
                closer := getattr(
                    self.loop,
                    "aclose",
                    None,
                )
            )
        ):
            await closer()

    def join(self) -> None:
        """Join the loop-runner thread."""
        self.loop_runner.join()

    def on_apply(
        self,
        target: AnyCallable | AnyCoroutine,
        args: tuple[Any, ...] = tuple(),
        kwargs: Optional[dict[str, Any]] = None,
        callback: Optional[AnyCallable | AnyCoroutine] = None,
        accept_callback: [AnyCallable | AnyCoroutine] = None,
        pid: Optional[int] = None,
        getpid: Callable[[], int] = os.getpid,
        propagate: tuple[AnyException, ...] = tuple(),
        monotonic: Callable[[], int] = time.monotonic,
        **_,
    ):
        """Apply function within pool context."""
        kwargs = kwargs or dict()
        propagate += (
            Exception,
            WorkerShutdown,
            WorkerTerminate,
        )

        if accept_callback:
            self.run(
                accept_callback,
                pid or getpid(),
                monotonic(),
            )

        try:
            ret = self.run(
                target,
                *args,
                **kwargs,
            )
        except propagate as error:
            raise error

        except BaseException as exc:
            try:
                reraise(
                    WorkerLostError,
                    WorkerLostError(repr(exc)),
                    sys.exc_info()[2],
                )
            except WorkerLostError:
                self.run(callback, ExceptionInfo())
        else:
            self.run(callback, ret)

    def terminate_job(self, pid, signal=None):
        """Terminate the specified job."""
        raise NotImplementedError

    def restart(self) -> None:
        """Restart the pool instance."""
        raise NotImplementedError



"""A custom implementation of Celery's built-in `build_tracer`
utility."""

# pylint: unused-argument

# Future Imports
from __future__ import annotations

# Standard Library Imports
import logging
import os
import time
from typing import (
    Any,
    Callable,
    Dict,
    FrozenSet,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
)

# Third-Party Imports
import celery
import celery.app
import celery.app.task
import celery.app.trace
import celery.canvas
import celery.exceptions
import celery.loaders
import celery.loaders.app
from celery.app.trace import AsyncResult, BackendGetMetaError, Context, EncodeError, ExceptionInfo, FAILURE, \
    gethostname, get_task_name, group, Ignore, IGNORED, IGNORE_STATES, info, InvalidTaskError, logger, LOG_IGNORED, \
    LOG_SUCCESS, Reject, REJECTED, report_internal_error, Retry, RETRY, safe_repr, saferepr, send_postrun, send_prerun, \
    send_success, _signal_internal_error, signals, STARTED, SUCCESS, successful_requests, task_has_custom, _task_stack, \
    traceback_clear, TraceInfo, trace_ok_t

# Package-Level Imports
from celery_aio_pool.types import AnyException

__all__ = ("build_async_tracer",)


# noinspection PyUnusedLocal
def build_async_tracer(
        name: str,
        task: Union[celery.Task, celery.local.PromiseProxy],
        loader: Optional[celery.loaders.app.AppLoader] = None,
        hostname: Optional[str] = None,
        store_errors: bool = True,
        Info: Type[TraceInfo] = TraceInfo,
        eager: bool = False,
        propagate: bool = False,
        app: Optional[celery.Celery] = None,
        monotonic: Callable[[], int] = time.monotonic,
        trace_ok_t: Type[trace_ok_t] = trace_ok_t,
        IGNORE_STATES: FrozenSet[str] = IGNORE_STATES) -> \
        Callable[[str, tuple[Any, ...], dict[str, Any], Any], trace_ok_t]:
    """Return a function that traces task execution.

    Catches all exceptions and updates result backend with the
    state and result.

    If the call was successful, it saves the result to the task result
    backend, and sets the task status to `"SUCCESS"`.

    If the call raises :exc:`~@Retry`, it extracts
    the original exception, uses that as the result and sets the task state
    to `"RETRY"`.

    If the call results in an exception, it saves the exception as the task
    result, and sets the task state to `"FAILURE"`.

    Return a function that takes the following arguments:

        uuid: The id of the task.
        args: List of positional args to pass on to the function.
        kwargs: Keyword arguments mapping to pass on to the function.
        request: Request dict.
    """

    # pylint: disable=too-many-statements

    # If the task doesn't define a custom __call__ method
    # we optimize it away by simply calling the run method directly,
    # saving the extra method call and a line less in the stack trace.
    fun = task if task_has_custom(task, '__call__') else task.run

    loader = loader or app.loader
    ignore_result = task.ignore_result
    track_started = task.track_started
    track_started = not eager and (task.track_started and not ignore_result)

    # #6476
    if eager and not ignore_result and task.store_eager_result:
        publish_result = True
    else:
        publish_result = not eager and not ignore_result

    deduplicate_successful_tasks = ((app.conf.task_acks_late or task.acks_late)
                                    and app.conf.worker_deduplicate_successful_tasks
                                    and app.backend.persistent)

    hostname = hostname or gethostname()
    inherit_parent_priority = app.conf.task_inherit_parent_priority

    loader_task_init = loader.on_task_init
    loader_cleanup = loader.on_process_cleanup

    task_before_start = None
    task_on_success = None
    task_after_return = None
    if task_has_custom(task, 'before_start'):
        task_before_start = task.before_start
    if task_has_custom(task, 'on_success'):
        task_on_success = task.on_success
    if task_has_custom(task, 'after_return'):
        task_after_return = task.after_return

    pid = os.getpid()

    request_stack = task.request_stack
    push_request = request_stack.push
    pop_request = request_stack.pop
    push_task = _task_stack.push
    pop_task = _task_stack.pop
    _does_info = logger.isEnabledFor(logging.INFO)
    resultrepr_maxsize = task.resultrepr_maxsize

    prerun_receivers = signals.task_prerun.receivers
    postrun_receivers = signals.task_postrun.receivers
    success_receivers = signals.task_success.receivers

    # Third-Party Imports
    from celery import canvas

    # Package-Level Imports
    from celery_aio_pool.pool import AsyncIOPool

    signature = canvas.maybe_signature  # maybe_ does not clone if already

    # noinspection PyUnusedLocal
    def on_error(
            request: celery.app.task.Context,
            exc: AnyException,
            state: str = FAILURE,
            call_errbacks: bool = True) -> Tuple[Info, Any, Any, Any]:
        """Handle any errors raised by a `Task`'s execution."""
        if propagate:
            raise
        I = Info(state, exc)
        R = I.handle_error_state(
            task, request, eager=eager, call_errbacks=call_errbacks,
        )
        return I, R, I.state, I.retval

    def trace_task(
            uuid: str,
            args: Sequence[Any],
            kwargs: Dict[str, Any],
            request: Optional[Dict[str, Any]] = None) -> trace_ok_t:
        """Execute and trace a `Task`."""

        # R      - is the possibly prepared return value.
        # I      - is the Info object.
        # T      - runtime
        # Rstr   - textual representation of return value
        # retval - is the always unmodified return value.
        # state  - is the resulting task state.

        # This function is very long because we've unrolled all the calls
        # for performance reasons, and because the function is so long
        # we want the main variables (I, and R) to stand out visually from the
        # the rest of the variables, so breaking PEP8 is worth it ;)
        R = I = T = Rstr = retval = state = None
        task_request = None
        time_start = monotonic()
        try:
            try:
                callable(kwargs.items)
            except AttributeError:
                raise InvalidTaskError(
                    'Task keyword arguments is not a mapping')

            task_request = Context(request or {}, args=args,
                                   called_directly=False, kwargs=kwargs)

            # TODO(the-wondersmith): Investigate if there are any potential
            #                        side effects of performing this update
            task.request.update(task_request.__dict__)

            redelivered = (task_request.delivery_info
                           and task_request.delivery_info.get('redelivered', False))
            if deduplicate_successful_tasks and redelivered:
                if task_request.id in successful_requests:
                    return trace_ok_t(R, I, T, Rstr)
                r = AsyncResult(task_request.id, app=app)

                try:
                    state = r.state
                except BackendGetMetaError:
                    pass
                else:
                    if state == SUCCESS:
                        info(LOG_IGNORED, {
                            'id': task_request.id,
                            'name': get_task_name(task_request, name),
                            'description': 'Task already completed successfully.'
                        })
                        return trace_ok_t(R, I, T, Rstr)

            push_task(task)
            root_id = task_request.root_id or uuid
            task_priority = task_request.delivery_info.get('priority') if \
                inherit_parent_priority else None
            push_request(task_request)
            try:
                # -*- PRE -*-
                if prerun_receivers:
                    send_prerun(sender=task, task_id=uuid, task=task,
                                args=args, kwargs=kwargs)
                AsyncIOPool.run_in_pool(loader_task_init, uuid, task)
                if track_started:
                    task.backend.store_result(
                        uuid, {'pid': pid, 'hostname': hostname}, STARTED,
                        request=task_request,
                    )

                # -*- TRACE -*-
                try:
                    if task_before_start:
                        AsyncIOPool.run_in_pool(task_before_start, uuid, args,
                                                kwargs)

                    R = retval = AsyncIOPool.run_in_pool(fun, *args, **kwargs)
                    state = SUCCESS
                except Reject as exc:
                    I, R = Info(REJECTED, exc), ExceptionInfo(internal=True)
                    state, retval = I.state, I.retval
                    I.handle_reject(task, task_request)
                    traceback_clear(exc)
                except Ignore as exc:
                    I, R = Info(IGNORED, exc), ExceptionInfo(internal=True)
                    state, retval = I.state, I.retval
                    I.handle_ignore(task, task_request)
                    traceback_clear(exc)
                except Retry as exc:
                    I, R, state, retval = on_error(
                        task_request, exc, RETRY, call_errbacks=False)
                    traceback_clear(exc)
                except Exception as exc:
                    I, R, state, retval = on_error(task_request, exc)
                    traceback_clear(exc)
                except BaseException:
                    raise
                else:
                    try:
                        # callback tasks must be applied before the result is
                        # stored, so that result.children is populated.

                        # groups are called inline and will store trail
                        # separately, so need to call them separately
                        # so that the trail's not added multiple times :(
                        # (Issue #1936)
                        callbacks = task.request.callbacks
                        if callbacks:
                            if len(task.request.callbacks) > 1:
                                sigs, groups = [], []
                                for sig in callbacks:
                                    sig = signature(sig, app=app)
                                    if isinstance(sig, group):
                                        groups.append(sig)
                                    else:
                                        sigs.append(sig)
                                for group_ in groups:
                                    group_.apply_async(
                                        (retval,),
                                        parent_id=uuid, root_id=root_id,
                                        priority=task_priority
                                    )
                                if sigs:
                                    group(sigs, app=app).apply_async(
                                        (retval,),
                                        parent_id=uuid, root_id=root_id,
                                        priority=task_priority
                                    )
                            else:
                                signature(callbacks[0], app=app).apply_async(
                                    (retval,), parent_id=uuid, root_id=root_id,
                                    priority=task_priority
                                )

                        # execute first task in chain
                        chain = task_request.chain
                        if chain:
                            _chsig = signature(chain.pop(), app=app)
                            _chsig.apply_async(
                                (retval,), chain=chain,
                                parent_id=uuid, root_id=root_id,
                                priority=task_priority
                            )
                        task.backend.mark_as_done(
                            uuid, retval, task_request, publish_result,
                        )
                    except EncodeError as exc:
                        I, R, state, retval = on_error(task_request, exc)
                    else:
                        Rstr = saferepr(R, resultrepr_maxsize)
                        T = monotonic() - time_start
                        if task_on_success:
                            AsyncIOPool.run_in_pool(task_on_success, retval,
                                                    uuid, args, kwargs)
                        if success_receivers:
                            send_success(sender=task, result=retval)
                        if _does_info:
                            info(LOG_SUCCESS, {
                                'id': uuid,
                                'name': get_task_name(task_request, name),
                                'return_value': Rstr,
                                'runtime': T,
                                'args': safe_repr(args),
                                'kwargs': safe_repr(kwargs),
                            })

                # -* POST *-
                if state not in IGNORE_STATES:
                    if task_after_return:
                        AsyncIOPool.run_in_pool(task_after_return,
                            state, retval, uuid, args, kwargs, None,
                        )
            finally:
                try:
                    if postrun_receivers:
                        send_postrun(sender=task, task_id=uuid, task=task,
                                     args=args, kwargs=kwargs,
                                     retval=retval, state=state)
                finally:
                    pop_task()
                    pop_request()
                    if not eager:
                        try:
                            task.backend.process_cleanup()
                            AsyncIOPool.run_in_pool(loader_cleanup)
                        except (KeyboardInterrupt, SystemExit, MemoryError):
                            raise
                        except Exception as exc:
                            logger.error('Process cleanup failed: %r', exc,
                                         exc_info=True)
        except MemoryError:
            raise
        except Exception as exc:
            _signal_internal_error(task, uuid, args, kwargs, request, exc)
            if eager:
                raise
            R = report_internal_error(task, exc)
            if task_request is not None:
                I, _, _, _ = AsyncIOPool.run_in_pool(on_error, task_request,
                                                     exc)
        return trace_ok_t(R, I, T, Rstr)

    return trace_task
